// 
//
#include "Server.h"

#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <iostream.h>

#define	MIN(a,b) ((a)<(b)?(a):(b))

Server::Server( char const *host, int port ) : pos(0), pos_max(0), sock(-1)
{
   Open();
   SetPort(port);
   if ( SetServer(host) == OK ) Connect();
   else cerr << "Unknown Host: '" << host << "'" << endl;
}

Server::~Server() 
{
   Close();
}

int Server::Open(void)
{
   sock = socket(AF_INET, SOCK_STREAM, 0);
   if (sock == NOTOK) return NOTOK;

   int on = 1;
   setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on));
   server.sin_family = AF_INET;

   return OK;
}

int Server::Close()
{
   if (sock >= 0) {
      int ret = close(sock);
      sock = -1;
      return ret;
   }
   return NOTOK;
}

int Server::SetPort(int port)
{
   server.sin_port = htons(port);
   return OK;
}

int Server::SetServer(char const *name)
{
   struct hostent *hp;
   unsigned long addr;

   addr = inet_addr((char *)name);
   if (addr == ~0L) {
      hp = gethostbyname(name);
      if (hp == NULL) return NOTOK;
      memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length);
   } else {
      memcpy((char *)&server.sin_addr, (char *)&addr, sizeof(addr));
   }

   return OK;
}

unsigned long Server::current=0;
char *Server::currName = 0;
int Server::currPort = 80;
void Server::SetCurrentServer( char const *name )
{
   struct hostent *hp;

   if (currName) delete currName;
   currName = strdup(name);
   current = inet_addr((char *)name);
   if (current == ~0L) {
      hp = gethostbyname(name);
      if ( hp != 0 ) memcpy((char *)&current, (char *)hp->h_addr, hp->h_length);
   } 
}

int Server::IsCurrent( char const *name, int port )
{
   struct hostent *hp;
   unsigned long addr;
   
   if ( currPort != port ) return 0; 

   addr = inet_addr((char *)name);
   if (addr == ~0L) {
      hp = gethostbyname(name);
      if ( hp == 0 ) return 0;
      memcpy((char *)&addr, (char *)hp->h_addr, hp->h_length);
   } 
   return current == addr; 
}

int Server::Connect( int allow_EINTR )
{
   int status;

   for (;;) {
      status = connect(sock, (struct sockaddr *)&server, sizeof(server));
      if (status < 0 && errno == EINTR && !allow_EINTR) {
	 close(sock);
	 Open();
	 continue;
      }
      break;
   }
	
   if (status == 0 || errno == EALREADY || errno == EISCONN) return OK;

   close(sock);
   Open();

   return NOTOK;
}

int Server::Bind()
{
   if (bind(sock, (struct sockaddr *)&server, sizeof(server)) == NOTOK) return NOTOK;
   return OK;
}

int Server::Listen(int n)
{
   return listen(sock, n);
}

unsigned long gethostip(char *ip, int length)
{
   char hostname[100];
   if (gethostname(hostname, sizeof(hostname)) == NOTOK) return 0;

   struct hostent *ent = gethostbyname(hostname);
   if (!ent) return 0;

   struct in_addr addr;
   memcpy((char *) &addr.s_addr, ent->h_addr, sizeof(addr));
   if (ip) strncpy(ip, inet_ntoa(addr), length);
   return addr.s_addr;
}

int Server::ReadChar( void )
{
   if (pos >= pos_max) {
      pos_max = ReadPartial(buffer, sizeof(buffer));
      pos = 0;
      if (pos_max <= 0) return -1;
   }
   return buffer[pos++] & 0xff;
}

char *Server::ReadLine( char *buffer, int maxlength, char *terminator )
{
   char	*start = buffer;
   int termseq = 0;

   while (maxlength > 0) {

      int ch = ReadChar();
      if (ch < 0) {
	 //
	 // End of file reached.  If we still have stuff in the input buffer
	 // we need to return it first.  When we get called again, we will
	 // return NULL to let the caller know about the EOF condition.
	 //
	 if (buffer > start) break;
	 else return (char *) 0;

      } else if (terminator[termseq] && ch == terminator[termseq]) {
	 //
	 // Got one of the terminator characters.  We will not put it in the string
	 // but keep track of the fact that we have seen it.
	 //
	 termseq++;
	 if (!terminator[termseq]) break;

      } else {
	 *buffer++ = ch;
	 maxlength--;
      }
   }
   *buffer = '\0';

   return start;
}

int Server::WriteLine(char *str, char *eol)
{
   int n, nn;

   if ((n = Write(str)) < 0) return -1;
   if ((nn = Write(eol)) < 0) return -1;
   return n + nn;
}

int Server::Write(char *buffer, int length)
{
   int nleft, nwritten;

   if (length == -1) length = strlen(buffer);
   nleft = length;

   while (nleft > 0) {
      nwritten = WritePartial(buffer, nleft);
      if (nwritten < 0 && errno == EINTR) continue;
      if (nwritten <= 0) return nwritten;
      nleft -= nwritten;
      buffer += nwritten;
   }
   return length - nleft;
}

int Server::Read(char *buffer, int length)
{
   int nleft, nread;

   nleft = length;

   //
   // If there is data in our internal input buffer, use that first.
   //
   if (pos < pos_max) {
      int n = MIN(length, pos_max - pos);
      memcpy(buffer, &this->buffer[pos], n);
      pos += n;
      buffer += n;
      nleft -= n;
   }

   while (nleft > 0) {
      nread = ReadPartial(buffer, nleft);
      if (nread < 0 && errno == EINTR) continue;
      if (nread < 0) return -1;
      else if (nread == 0) break;
      nleft -= nread;
      buffer += nread;
   }
   return length - nleft;
}

int Server::ReadPartial(char *buffer, int maxlength)
{
   int count;

   do {
      count = ::read(sock, buffer, maxlength);
   } while (count < 0 && errno == EINTR);

   return count;
}

int Server::WritePartial(char *buffer, int maxlength)
{
   int count;

   do {
      count = ::write(sock, buffer, maxlength);
   } while (count < 0 && errno == EINTR);

   return count;
}
